"""
This submodule contains a fake camera implementation that can be used for
testing purposes. All instances of the FakeCamera class will share the
same class structure so type-checking will work as expected. Similarly,
the FakeCamera class will be a subclass of RequestMachinery so it can be
used as a drop-in replacement for a real camera in basically every way.
"""
import time
from concurrent.futures import Future
from threading import Event, Thread
from typing import Any, Dict, Tuple

import libcamera
import numpy as np

from scicamera.actions import RequestMachinery
from scicamera.configuration import CameraConfig, StreamConfig
from scicamera.controls import Controls
from scicamera.info import CameraInfo
from scicamera.request import CompletedRequest
from scicamera.typing import TypedFuture

FAKE_SIZE = (320, 240)
FAKE_FORMAT = "RGB888"
FAKE_CHANNELS = 3
FAKE_STRIDE = FAKE_CHANNELS * FAKE_SIZE[0]


def make_fake_image(shape: Tuple[int, int]):
    """Make a image buffer in the style the cameras might return

    This is intended to be a ``RGB888`` encoding, so the size/shape
    isn't what most people are used to for images in ``numpy`` land.
    """
    w, h = shape
    img = np.zeros((h, w, FAKE_CHANNELS), dtype=np.uint8)

    img[:, : w // 3, 0] = 255
    img[:, w // 3 : 2 * w // 3, 1] = 255
    img[:, 2 * w // 3 :, 2] = 255
    return img


class FakeCompletedRequest(CompletedRequest):
    def __init__(self, config: CameraConfig, metadata: Dict[str, Any]):
        self.config = config
        self.completion_time = time.time()
        self._metadata = metadata
        self._metadata["SensorTimestamp"] = int(
            (self.completion_time - 1.0) * 1_000_000_000
        )

    def acquire(self):
        pass

    def release(self):
        pass

    def get_camera_config(self) -> CameraConfig:
        """Fetch the configuration for the named stream."""
        return self.config

    def get_buffer(self, name: str) -> np.ndarray:
        """Make a 1d numpy array from the named stream's buffer."""
        size = self.config.get_stream_config(name).size
        return make_fake_image(size).flatten()

    def get_metadata(self) -> Dict[str, Any]:
        """Fetch the metadata corresponding to this completed request."""
        return self._metadata


class FakeCamera(RequestMachinery):
    def __init__(self, camera_num: int = 0, tuning=None) -> None:
        super().__init__()
        self._t = Thread(target=lambda: None, daemon=True)
        self._t.start()
        self._t.join()
        self._abort = Event()

        self.sensor_resolution = FAKE_SIZE
        self.sensor_format = FAKE_FORMAT
        self.camera_config = None
        self.camera_ctrl_info = {
            "AeEnable": 0,
            "AnalogueGain": 1.0,
            "DigitalGain": 1.0,
            "AwbEnable": 0,
            "ColourCorrectionMatrix": [1, 0, 0, 0, 1, 0, 0, 0, 1],
            "ColourGains": (2.5220680236816406, 1.8971731662750244),
            "ColourTemperature": 4000,
            "ExposureTime": 10000000,
            "FrameDurationLimits": (33333, 33333),
            "NoiseReductionMode": 0,
        }
        self.controls = Controls(self, self.camera_ctrl_info)

        self.config = CameraConfig(
            self,
            "still",
            1,
            controls=self.controls,
            transform=libcamera.Transform(),
            color_space=libcamera.ColorSpace.Sycc(),
            main=StreamConfig(size=FAKE_SIZE, format=FAKE_FORMAT, stride=FAKE_STRIDE),
        )

    def _run(self):
        while not self._abort.wait(self.controls.FrameDurationLimits[0] / 1000000):
            metadata = self.controls.make_dict()

            metadata.update(
                {
                    "AeLocked": False,
                    "FocusFoM": 93,
                    "FrameDuration": 24994,
                    "Lux": 330.6990051269531,
                }
            )
            request = FakeCompletedRequest(self.config, metadata)
            self.add_completed_request(request)
            self.process_requests()

    def configure(self, config: CameraConfig) -> None:
        self.camera_config = config
        self.controls.set_controls(config.controls)

    @property
    def info(self) -> CameraInfo:
        return CameraInfo(
            id="fake",
            model="Fake Camera",
            size=FAKE_SIZE,
            location="Nowhere",
            rotation=1,
        )

    @property
    def camera_controls(self):
        return {}

    def start_runloop(self):
        pass

    def stop_runloop(self):
        pass

    def start(self) -> None:
        self._t = Thread(target=self._run, daemon=True)
        self._abort.clear()
        self._t.start()

    def stop(self) -> None:
        self._abort.set()
        self._t.join()

    def close(self) -> None:
        if self._t.is_alive():
            self.stop()

    def switch_mode(self, camera_config: CameraConfig) -> TypedFuture[CameraConfig]:
        self.configure(camera_config)
        future = Future()
        future.set_result(camera_config)
        return future

    # TODO(meawoppl) - Kill methods below here
    def set_controls(self, controls: Dict[str, Any]) -> None:
        self.controls.set_controls(controls)

    def camera_configuration(self) -> CameraConfig:
        return self.camera_config



import importlib.metadata

import libcamera

from scicamera.camera import Camera, CameraManager
from scicamera.configuration import CameraConfig, StreamConfig
from scicamera.controls import Controls
from scicamera.fake import FakeCamera
from scicamera.info import CameraInfo
from scicamera.lc_helpers import libcamera_color_spaces_eq, libcamera_transforms_eq

# NOTE(meawoppl) - ugleeee monkey patch. Kill the below VV
libcamera.Transform.__repr__ = libcamera.Transform.__str__
libcamera.Transform.__eq__ = libcamera_transforms_eq

libcamera.ColorSpace.__repr__ = libcamera.ColorSpace.__str__
libcamera.ColorSpace.__eq__ = libcamera_color_spaces_eq


__all__ = [
    "Camera",
    "CameraConfig",
    "CameraInfo",
    "CameraManager",
    "CompletedRequest",
    "Controls",
    "FakeCamera",
    "StreamConfig",
]
try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except ModuleNotFoundError:
    __version__ = "0.0.0-local"



from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from scicamera.request import CompletedRequest


@dataclass
class CameraFrame:
    array: np.ndarray
    """The image data as a numpy array."""

    controls: dict
    """The configuration used to capture the image."""

    metadata: dict
    """The metadata associated with the image."""

    @classmethod
    def from_request(cls, name: str, request: CompletedRequest) -> CameraFrame:
        """Create a CameraFrame from a CompletedRequest."""
        return cls(
            array=request.make_array(name),
            controls=request.config.controls.make_dict(),
            metadata=request.get_metadata(),
        )



import math
from logging import getLogger

from scicamera.camera import Camera
from scicamera.request import CompletedRequest

_log = getLogger(__name__)


def calibrate_camera_offset(
    camera: Camera,
    n_frames: int = 100,
) -> float:
    """Calibrate the ``SensorTimestamp`` wrt to the epoch time.

    Returns the number of nanoseconds you should add to the camera
    ``SensorTimestamp`` to get the epoch time in nanoseconds.
    """
    deltas = []

    def _capture_timing_callback(request: CompletedRequest):
        # This is the time the request was handed to python
        epoch_nanos = int(request.completion_time * 1_000_000_000)
        # This is the time the "sensor" reports `ktime_get_ns()` in the kernel
        sensor_nanos = request.get_metadata()["SensorTimestamp"]
        deltas.append(epoch_nanos - sensor_nanos)

    # Make sure the above callback happens at least `n_frames` times
    camera.add_request_callback(_capture_timing_callback)
    camera.discard_frames(n_frames).result()
    camera.remove_request_callback(_capture_timing_callback)

    # NB: This segment relies on python's integer size growth
    offset = sum(deltas) / len(deltas)
    diffs = sum([(d - offset) ** 2 for d in deltas])
    stdev = math.sqrt(diffs / (len(deltas) - 1))

    _log.warning(f"Camera offset: {offset} ± {stdev}")
    return offset



import logging
import re
import sys
from concurrent.futures._base import TimeoutError as FuturesTimeoutError
from typing import Iterable

import pytest

from scicamera import Camera, FakeCamera

_log = logging.getLogger(__name__)


def mature_after_frames_or_timeout(
    camera: Camera, n_frames: int = 2, timeout_seconds=5
):
    """Return a future that will be mature after n_frames or 2 seconds.

    Raises: TimeoutError if it takes too long.
    """
    try:
        camera.discard_frames(n_frames).result(timeout_seconds)
    except FuturesTimeoutError as e:
        raise TimeoutError("Timed out waiting for camera to mature") from e


def requires_camera_model(camera: Camera, model_pattern: str, allow_fake: bool = True):
    if isinstance(camera, FakeCamera) and allow_fake:
        return
    model_name = camera.info.model
    if not re.match(model_pattern, model_name):
        _log.warning("Closing camera in fixture.")
        camera.close()
        pytest.skip(
            f"Skipping test, camera model {model_name} does not match {model_pattern}"
        )


def requires_controls(camera: Camera, controls: Iterable[str]):
    """Decorator to skip tests if the camera does not support the given controls."""
    # NOTE(meawoppl) - This should be a pytest.skip(), but because of the way
    # tests are run in subprocesses, pytest.skip() doesn't work... TODO FIXME

    available_controls = camera.controls.available_control_names()
    missing_controls = set(controls) - available_controls

    if missing_controls:
        print("Skipping test, missing controls:", missing_controls)
        sys.exit(0)



import math
from typing import Tuple

import numpy as np

from scicamera.sensor_format import SensorFormat

YUV_FORMATS = {"NV21", "NV12", "YUV420", "YVU420", "YVYU", "YUYV", "UYVY", "VYUY"}

RGB_FORMATS = {"BGR888", "RGB888", "XBGR8888", "XRGB8888"}

BAYER_FORMATS = {
    "SBGGR8",
    "SGBRG8",
    "SGRBG8",
    "SRGGB8",
    "SBGGR10",
    "SGBRG10",
    "SGRBG10",
    "SRGGB10",
    "SBGGR10_CSI2P",
    "SGBRG10_CSI2P",
    "SGRBG10_CSI2P",
    "SRGGB10_CSI2P",
    "SBGGR12",
    "SGBRG12",
    "SGRBG12",
    "SRGGB12",
    "SBGGR12_CSI2P",
    "SGBRG12_CSI2P",
    "SGRBG12_CSI2P",
    "SRGGB12_CSI2P",
}

MONO_FORMATS = {"R8", "R10", "R12", "R8_CSI2P", "R10_CSI2P", "R12_CSI2P"}

ALL_FORMATS = YUV_FORMATS | RGB_FORMATS | BAYER_FORMATS | MONO_FORMATS


def is_YUV(fmt: str) -> bool:
    return fmt in YUV_FORMATS


def is_RGB(fmt: str) -> bool:
    return fmt in RGB_FORMATS


def is_Bayer(fmt: str) -> bool:
    return fmt in BAYER_FORMATS


def is_mono(fmt: str) -> bool:
    return fmt in MONO_FORMATS


def is_raw(fmt: str) -> bool:
    return is_Bayer(fmt) or is_mono(fmt)


def is_format_valid(fmt: str) -> bool:
    return fmt in ALL_FORMATS


def assert_format_valid(fmt: str) -> None:
    if not is_format_valid(fmt):
        raise ValueError(f"Invalid format: {fmt}. Valid formats are: {ALL_FORMATS}")


def _assert_is_byte_array(array: np.ndarray) -> None:
    assert array.dtype == np.uint8, "Raw unpack only accepts uint8 arrays"
    assert array.ndim == 1, "Unpack raw only accepts flat arrays"


# NOTE(meawoppl) - the below implementations are a bit memory inefficient when it comes
# to deserialization of the 10/12 bit arrays, as it will overallocated by 1/5 and 1/3
# respectively. This is a tradeoff for simplicity of implementation using numpy.
# The approach breaks down to the following:
# - Compute the number of bytes needed at which the data realigns itself to the next byte boundary
#   - for 10 bit (4 pixels) - 40 bits, 5 bytes
#   - fot 12 bit (2 pixels) - 24 bits, 3 bytes
# - Unspool things into alignment blocks (0th axis) and realigned stuff within the blocks
# - Flatten the index space downward, and trim the tail of the array away (assumed extra bits)


def _unpack_10bit(array: np.ndarray) -> np.ndarray:
    original_len = array.size
    array16 = array.reshape((-1, 5)).astype(np.uint16)

    unpacked_data = np.zeros((array16.shape[0], 4), dtype=np.uint16)
    # fmt: off
    unpacked_data[:, 0] = ((array16[:, 0] << 2) | (array16[:, 1] >> 6)) & 0x3FF
    unpacked_data[:, 1] = ((array16[:, 1] << 4) | (array16[:, 2] >> 4)) & 0x3FF
    unpacked_data[:, 2] = ((array16[:, 2] << 6) | (array16[:, 3] >> 2)) & 0x3FF
    unpacked_data[:, 3] = ((array16[:, 3] << 8) | (array16[:, 4]     )) & 0x3FF
    # fmt: on

    return unpacked_data.ravel()[: original_len * 4 // 5]


def _unpack_12bit(array: np.ndarray) -> np.ndarray:
    original_len = array.size
    array16 = array.reshape((-1, 3)).astype(np.uint16)

    unpacked_data = np.zeros((array16.shape[0], 2), dtype=np.uint16)
    # fmt: off
    unpacked_data[:, 0] = ((array16[:, 0] << 4) | ((array16[:, 2] >> 0) & 0xF)) & 0xFFF
    unpacked_data[:, 1] = ((array16[:, 1] << 4) | ((array16[:, 2] >> 4) & 0xF)) & 0xFFF
    # fmt: on

    return unpacked_data.ravel()[: original_len * 2 // 3]


def round_up_to_multiple(value: int, multiple: int) -> int:
    return math.ceil(value / multiple) * multiple


def unpack_csi_padded(
    raw: np.ndarray, pixel_shape: Tuple[int, int], fmt: SensorFormat
) -> np.ndarray:
    """
    Args:
        raw: Flat uint8 array of bytes from the camera
        pixel_shape: the shape of the image in pixels (height, width)
        fmt:

    """
    _assert_is_byte_array(raw)
    assert fmt.packing == "CSI2P", "This method only treats CSI2P packing"

    # The Unicam (MIPI CSI?) series of sensors pad each row up to the nearest multiple of 32 bytes.
    # This is somewhat annoying to deal with, as we need also to also keep alignment
    # with the 10/12 bit packing.

    # 1. Compute the smallest nominal np array shape which meets the bit-alignment requirements
    row_bit_length = fmt.bit_depth * pixel_shape[1]
    row_byte_length = round_up_to_multiple(row_bit_length, 8) // 8
    row_byte_length_padded = round_up_to_multiple(row_byte_length, 32)

    expected_data_length = row_byte_length_padded * pixel_shape[0]
    if raw.size != expected_data_length:
        raise ValueError(
            f"Raw data size {raw.size} does not match expected size {expected_data_length}"
        )

    n_rows = raw.size // row_byte_length_padded
    raw_rows = raw.reshape((-1, row_byte_length_padded))

    if fmt.bit_depth == 10:
        aligned_row_length = round_up_to_multiple(row_byte_length_padded, 5)
    if fmt.bit_depth == 12:
        aligned_row_length = round_up_to_multiple(row_byte_length_padded, 3)

    # 2. Copy the raw data into that array
    aligned_for_unpacking = np.zeros(
        (raw_rows.shape[0], aligned_row_length), dtype=np.uint8
    )
    aligned_for_unpacking[:, :row_byte_length_padded] = raw_rows

    # 3. Flatten and convert the byte-array into a uint16 array
    if fmt.bit_depth == 10:
        unpacked = _unpack_10bit(aligned_for_unpacking.ravel())
    if fmt.bit_depth == 12:
        unpacked = _unpack_12bit(aligned_for_unpacking.ravel())

    # 4. reshape the uint16 array into the desired shape
    unpacked_padded = unpacked.reshape((n_rows, -1))

    # 5. trim the tails of the horizontal lines away to the correct shape
    trimmed = unpacked_padded[:, : pixel_shape[1]]
    return trimmed


def unpack_raw(
    raw: np.ndarray, pixel_shape: Tuple[int, int], fmt: SensorFormat
) -> np.ndarray:
    """
    This converts a raw numpy byte array (flat, uint8) into a 2d numpy array of `pixel_shape`
    and dtype baed on SensorFormat. Note that in most formats this will still be a bayered image.
    """
    _assert_is_byte_array(raw)

    # TODO(meawoppl) - add other packed formats here
    if fmt.packing == "CSI2P":
        return unpack_csi_padded(raw, pixel_shape, fmt)
    else:
        raise RuntimeError(f"Unsupported bit raw format: {fmt}")



from __future__ import annotations

import io
import mmap
import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import Future
from dataclasses import dataclass, field
from functools import partial
from logging import getLogger
from typing import Any, Callable, Dict

import numpy as np
from PIL import Image

import scicamera.formats as formats
from scicamera import formats
from scicamera.configuration import CameraConfig
from scicamera.formats import unpack_raw
from scicamera.lc_helpers import lc_unpack
from scicamera.sensor_format import SensorFormat

_log = getLogger(__name__)


class MappedBuffer:
    def __init__(self, lc_buffer):
        self.__fb = lc_buffer

    def __enter__(self):
        # Check if the buffer is contiguous and find the total length.
        fd = self.__fb.planes[0].fd
        planes_metadata = self.__fb.metadata.planes
        buflen = 0
        for p, p_metadata in zip(self.__fb.planes, planes_metadata):
            # bytes_used is the same as p.length for regular frames, but correctly reflects
            # the compressed image size for MJPEG cameras.
            buflen = buflen + p_metadata.bytes_used
            if fd != p.fd:
                raise RuntimeError("_MappedBuffer: Cannot map non-contiguous buffer!")

        self.__mm = mmap.mmap(
            fd, buflen, mmap.MAP_SHARED, mmap.PROT_READ | mmap.PROT_WRITE
        )
        return self.__mm

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.__mm is not None:
            self.__mm.close()


class AbstractCompletedRequest(ABC):
    @abstractmethod
    def get_camera_config(self) -> CameraConfig:
        raise NotImplementedError()

    @abstractmethod
    def get_buffer(self, name: str) -> np.ndarray:
        raise NotImplementedError()

    @abstractmethod
    def get_metadata(self) -> Dict[str, Any]:
        raise NotImplementedError()

    def make_array(self, name: str) -> np.ndarray:
        """Make a 2d numpy array from the named stream's buffer."""
        config = self.get_camera_config()
        stream_cfg = config.get_stream_config(name)
        w, h = stream_cfg.size
        stride = stream_cfg.stride
        fmt = stream_cfg.format

        array = self.get_buffer(name)

        # Turning the 1d array into a 2d image-like array only works if the
        # image stride (which is in bytes) is a whole number of pixels. Even
        # then, if they don't match exactly you will get "padding" down the RHS.
        # Working around this requires another expensive copy of all the data.
        if fmt in ("BGR888", "RGB888"):
            if stride != w * 3:
                array = array.reshape((h, stride))
                array = np.asarray(array[:, : w * 3], order="C")
            image = array.reshape((h, w, 3))
        elif fmt in ("XBGR8888", "XRGB8888"):
            if stride != w * 4:
                array = array.reshape((h, stride))
                array = np.asarray(array[:, : w * 4], order="C")
            image = array.reshape((h, w, 4))
        elif fmt in ("YUV420", "YVU420"):
            # Returning YUV420 as an image of 50% greater height (the extra bit continaing
            # the U/V data) is useful because OpenCV can convert it to RGB for us quite
            # efficiently. We leave any packing in there, however, as it would be easier
            # to remove that after conversion to RGB (if that's what the caller does).
            image = array.reshape((h * 3 // 2, stride))
        elif fmt in ("YUYV", "YVYU", "UYVY", "VYUY"):
            # These dimensions seem a bit strange, but mean that
            # cv2.cvtColor(image, cv2.COLOR_YUV2BGR_YUYV) will convert directly to RGB.
            image = array.reshape(h, stride // 2, 2)
        elif fmt == "MJPEG":
            image = np.array(Image.open(io.BytesIO(array)))
        elif formats.is_raw(fmt):
            image = unpack_raw(array, (h, w), SensorFormat(fmt))
        else:
            raise RuntimeError("Format " + config.format + " not supported")
        return image

    def make_image(self, name: str) -> Image.Image:
        """Make a PIL image from the named stream's buffer."""
        fmt = self.get_camera_config().get_stream_config(name).format
        if fmt == "MJPEG":
            buffer = self.get_buffer(name)
            return Image.open(io.BytesIO(buffer))

        rgb = self.make_array(name)
        mode_lookup = {
            "RGB888": "BGR",
            "BGR888": "RGB",
            "XBGR8888": "RGBA",
            "XRGB8888": "BGRX",
        }
        if fmt not in mode_lookup:
            raise RuntimeError(f"Stream format {fmt} not supported for PIL images")
        mode = mode_lookup[fmt]
        return Image.frombuffer(
            "RGB", (rgb.shape[1], rgb.shape[0]), rgb, "raw", mode, 0, 1
        )


# TODO(meawoppl) - Make Completed Requests only exist inside of a context manager
# This remove all the bizzare locking and reference counting we are doing here manually
class CompletedRequest(AbstractCompletedRequest):
    def __init__(
        self,
        lc_request,
        config: CameraConfig,
        stream_map: Dict[str, Any],
        cleanup: Callable[[], None],
    ):
        self.completion_time = time.time()
        self.request = lc_request
        self.ref_count = 1
        self.lock = threading.Lock()
        self.config = config
        self.cleanup = cleanup
        self.stream_map = stream_map

    def acquire(self):
        """Acquire a reference to this completed request, which stops it being recycled back to
        the camera system.
        """
        with self.lock:
            if self.ref_count == 0:
                raise RuntimeError("CompletedRequest: acquiring lock with ref_count 0")
            self.ref_count += 1

    def release(self):
        """Release this completed frame back to the camera system (once its reference count
        reaches zero).
        """
        with self.lock:
            self.ref_count -= 1
            if self.ref_count < 0:
                raise RuntimeError("CompletedRequest: lock now has negative ref_count")

            if self.ref_count > 0:
                return

            self.cleanup()
            self.request = None

    def get_camera_config(self) -> CameraConfig:
        """Fetch the configuration for the named stream."""
        return self.config

    def get_buffer(self, name: str) -> np.ndarray:
        """Make a 1d numpy array from the named stream's buffer."""
        stream = self.stream_map[name]
        buffer = self.request.buffers[stream]
        with MappedBuffer(buffer) as b:
            return np.array(b, dtype=np.uint8)

    def get_metadata(self) -> Dict[str, Any]:
        """Fetch the metadata corresponding to this completed request."""
        return lc_unpack(self.request.metadata)


@dataclass
class LoopTask:
    call: Callable[[CompletedRequest], Any] | callable[[], Any]

    needs_request: bool = True

    future: Future = field(init=False, default_factory=Future)

    @classmethod
    def with_request(cls, call, *args):
        return cls(call=partial(call, *args), needs_request=True)

    @classmethod
    def without_request(cls, call, *args):
        return cls(call=partial(call, *args), needs_request=False)

    def __post_init__(self):
        self.future.set_running_or_notify_cancel()



from __future__ import annotations

from concurrent.futures import Future
from typing import Any, Callable, Generic, TypeVar

T = TypeVar("T")


class TypedFuture(Future, Generic[T]):
    def add_done_callback(
        self: TypedFuture, fn: Callable[[TypedFuture[T]], Any]
    ) -> None:
        ...

    def cancel(self: TypedFuture) -> bool:
        ...

    def cancelled(self: TypedFuture) -> bool:
        ...

    def done(self: TypedFuture) -> bool:
        ...

    def exception(self):
        ...

    def result(self: TypedFuture[T], timeout: float | None = ...) -> T:
        ...

    def running(self: TypedFuture[T]) -> bool:
        ...

    def set_result(self: TypedFuture[T], result: T) -> None:
        ...

    def set_running_or_notify_cancel(self: TypedFuture[T]) -> bool:
        ...

    def set_exception(self: TypedFuture[T], exception: Exception) -> None:
        ...



from __future__ import annotations

from typing import Set

from libcamera import ControlType, Rectangle, Size


def _framerates_to_durations(framerates):
    if not isinstance(framerates, (tuple, list)):
        framerates = (framerates, framerates)
    return (int(1000000 / framerates[1]), int(1000000 / framerates[0]))


def _durations_to_framerates(durations):
    if durations[0] == durations[1]:
        return 1000000 / durations[0]
    return (1000000 / durations[1], 1000000 / durations[0])


class Controls:
    def __init__(self, camera, controls={}):
        self._camera = camera
        self._controls = []
        self.set_controls(controls)

    def available_control_names(self) -> Set[str]:
        """Returns a set of all available control names"""
        return set(self._camera.camera_ctrl_info.keys())

    def __setattr__(self, name: str, value):
        if not name.startswith("_"):
            if name not in self.available_control_names():
                raise RuntimeError(f"Control {name} is not advertised by libcamera")
            self._controls.append(name)
        self.__dict__[name] = value

    def __repr__(self):
        return f"<Controls: {self.make_dict()}>"

    def set_controls(self, controls: dict | Controls):
        if isinstance(controls, dict):
            for k, v in controls.items():
                self.__setattr__(k, v)
        elif isinstance(controls, Controls):
            for k in controls._controls:
                v = controls.__dict__[k]
                self.__setattr__(k, v)
        else:
            raise RuntimeError(f"Cannot update controls with {type(controls)} type")

    def get_frame_rate(self) -> float:
        return _durations_to_framerates(self.FrameDurationLimits)

    def set_frame_rate(self, value: float):
        self.FrameDurationLimits = _framerates_to_durations(value)

    def get_libcamera_controls(self) -> dict:
        libcamera_controls = {}
        for k in self._controls:
            v = self.__dict__[k]
            id = self._camera.camera_ctrl_info[k][0]
            if id.type == ControlType.Rectangle:
                v = Rectangle(*v)
            elif id.type == ControlType.Size:
                v = Size(*v)
            libcamera_controls[id] = v
        return libcamera_controls

    def make_dict(self):
        dict_ = {}
        for k in self._controls:
            v = self.__dict__[k]
            dict_[k] = v
        return dict_



from abc import ABC, abstractmethod
from collections import deque
from concurrent.futures import Future
from logging import getLogger
from threading import Condition, Event, Thread
from typing import Any, Callable, Deque, Dict, List, Optional, Tuple

import numpy as np
from PIL import Image

from scicamera.configuration import CameraConfig
from scicamera.frame import CameraFrame
from scicamera.request import CompletedRequest, LoopTask
from scicamera.typing import TypedFuture

_log = getLogger(__name__)


class RequestMachinery(ABC):
    """RequestMachinery is a helper class for the Camera class."""

    def __init__(self) -> None:
        self._requests: Deque[CompletedRequest] = deque()
        self._request_callbacks: List[Callable[[CompletedRequest], None]] = []
        self._task_deque: Deque[LoopTask] = deque()

        self._runloop_cond = Condition()
        self._runloop_abort = Event()
        self._runloop_thread = Thread(target=lambda: 0, daemon=True)
        self._runloop_thread.start()
        self._runloop_thread.join()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add_request_callback(self, callback: Callable[[CompletedRequest], None]):
        """Add a callback to be called when every request completes.

        Note that the request is only valid within the callback, and will be
        deallocated after the callback returns.

        :param callback: The callback to be called
        :type callback: Callable[[CompletedRequest], None]
        """
        self._request_callbacks.append(callback)

    def remove_request_callback(self, callback: Callable[[CompletedRequest], None]):
        """Remove a callback previously added with add_request_callback.

        :param callback: The callback to be removed
        :type callback: Callable[[CompletedRequest], None]
        """
        self._request_callbacks.remove(callback)

    def add_completed_request(self, request: CompletedRequest) -> None:
        self._requests.append(request)

        with self._runloop_cond:
            self._runloop_cond.notify()

    def _runloop(self) -> None:
        while True:
            with self._runloop_cond:
                self._runloop_cond.wait(timeout=0.05)

            if self._runloop_abort.is_set():
                break

            if len(self._requests) > 0:
                self.process_requests()

    def start_runloop(self) -> None:
        """
        Start the preview loop.
        """
        if self._runloop_thread.is_alive():
            raise RuntimeError("Runloop thread already running?")

        self._runloop_abort.clear()
        self._runloop_thread = Thread(target=self._runloop, daemon=True)
        self._runloop_thread.start()

    def is_runloop_running(self):
        return self._runloop_thread.is_alive()

    def stop_runloop(self) -> None:
        """Stop preview

        :raises RuntimeError: Unable to stop preview
        """
        self._runloop_abort.set()
        with self._runloop_cond:
            self._runloop_cond.notify()
        self._runloop_thread.join()

    def has_requests(self) -> bool:
        return len(self._requests) > 0

    def process_requests(self) -> None:
        # Safe copy and pop off all requests
        requests = list(self._requests)
        for _ in requests:
            self._requests.popleft()

        req_idx = 0
        while len(self._task_deque) and (req_idx < len(requests)):
            task = self._task_deque.popleft()
            _log.debug(f"Begin LoopTask Execution: {task.call}")
            try:
                if task.needs_request:
                    req = requests[req_idx]
                    req_idx += 1
                    task.future.set_result(task.call(req))
                else:
                    task.future.set_result(task.call())
            except Exception as e:
                _log.warning(f"Error in LoopTask {task.call}: {e}")
                task.future.set_exception(e)
            _log.debug(f"End LoopTask Execution: {task.call}")

        for request in requests:
            for callback in self._request_callbacks:
                try:
                    callback(request)
                except Exception as e:
                    _log.error(f"Error in request callback ({callback}): {e}")

        for req in requests:
            req.release()

    def _dispatch_loop_tasks(
        self, *args: LoopTask, config: Optional[dict] = None
    ) -> List[Future]:
        """The main thread should use this to dispatch a number of operations for the event
        loop to perform. The event loop will execute them in order, and return a list of
        futures which mature at the time the corresponding operation completes.
        """

        if config is None:
            tasks = args
        else:
            previous_config = self.camera_config
            # FIXME: the discarded request enough for test cases, but the correct
            # way to flag this is with the request.cookie, but that is currently
            # used to route between cameras. Fixable, but independent issue for now.
            tasks = (
                [
                    LoopTask.without_request(self._switch_mode, config),
                    LoopTask.with_request(self._discard_request),
                ]
                + list(args)
                + [
                    LoopTask.without_request(self._switch_mode, previous_config),
                ]
            )
        self._task_deque.extend(tasks)
        # Note that the below strips the config changes
        return [task.future for task in args]

    def _discard_request(self, request: CompletedRequest) -> None:
        pass

    def discard_frames(self, n_frames: int) -> TypedFuture[None]:
        """Discard the next ``n_frames`` in the queue."""
        return self._dispatch_loop_tasks(
            *[LoopTask.with_request(self._discard_request) for _ in range(n_frames)]
        )[-1]

    def _switch_mode(self, camera_config: CameraConfig):
        self._stop()
        self.configure(camera_config)
        self._start()
        return self.camera_config

    def switch_mode(self, camera_config: CameraConfig) -> TypedFuture[CameraConfig]:
        """Switch the camera into another mode given by the camera_config."""
        return self._dispatch_loop_tasks(
            LoopTask.without_request(self._switch_mode, camera_config)
        )[0]

    def _capture_file(
        self, name: str, file_output, format, request: CompletedRequest
    ) -> dict:
        request.make_image(name).convert("RGB").save(file_output, format=format)
        return request.get_metadata()

    def capture_file(
        self, file_output, name: str = "main", format=None, config=None
    ) -> TypedFuture[dict]:
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_file, name, file_output, format),
            config=config,
        )[0]

    def _capture_request(self, request: CompletedRequest):
        request.acquire()
        return request

    def capture_request(
        self, config: Optional[dict] = None
    ) -> TypedFuture[CompletedRequest]:
        """Fetch the next completed request from the camera system. You will be holding a
        reference to this request so you must release it again to return it to the camera system.
        """
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_request), config=config
        )[0]

    def _capture_metadata(self, request: CompletedRequest):
        return request.get_metadata()

    def capture_metadata(self, config: Optional[dict] = None) -> Future:
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_metadata), config=config
        )[0]

    # Buffer Capture Methods
    def _capture_buffer(self, name: str, request: CompletedRequest):
        return request.get_buffer(name)

    def capture_buffer(self, name="main", config: dict = None) -> Future[np.ndarray]:
        """Make a 1d numpy array from the next frame in the named stream."""
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_buffer, name), config=config
        )[0]

    # Buffers and metadata
    def _capture_buffers_and_metadata(
        self, names: List[str], request: CompletedRequest
    ) -> Tuple[List[np.ndarray], dict]:
        return ([request.get_buffer(name) for name in names], request.get_metadata())

    def capture_buffers_and_metadata(
        self, names=["main"]
    ) -> TypedFuture[Tuple[List[np.ndarray], dict]]:
        """Make a 1d numpy array from the next frame for each of the named streams."""
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_buffers_and_metadata, names)
        )[0]

    # Array Capture Methods
    def _capture_array(self, name: str, request: CompletedRequest):
        return request.make_array(name)

    def capture_array(
        self, name: str = "main", config: Optional[dict] = None
    ) -> Future[np.ndarray]:
        """Make a 2d image from the next frame in the named stream."""
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_array, name), config=config
        )[0]

    def _capture_arrays_and_metadata(
        self, names: List[str], request: CompletedRequest
    ) -> Tuple[List[np.ndarray], Dict[str, Any]]:
        return ([request.make_array(name) for name in names], request.get_metadata())

    def capture_arrays_and_metadata(
        self, names: List[str] = ["main"]
    ) -> TypedFuture[Tuple[List[np.ndarray], Dict[str, Any]]]:
        """Make 2d image arrays from the next frames in the named streams."""
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_arrays_and_metadata, names)
        )[0]

    def _capture_image(self, name: str, request: CompletedRequest) -> Image.Image:
        return request.make_image(name)

    def capture_image(
        self, name: str = "main", config: Optional[dict] = None
    ) -> TypedFuture[Image.Image]:
        """Make a PIL image from the next frame in the named stream.

        :param name: Stream name, defaults to "main"
        :type name: str, optional
        :param wait: Wait for the event loop to finish an operation and signal us, defaults to True
        :type wait: bool, optional
        :param signal_function: Callback, defaults to None
        :type signal_function: function, optional
        :return: PIL Image
        :rtype: Image
        """
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_image, name),
            config=config,
        )[0]

    def _capture_frame(self, name: str, request: CompletedRequest) -> CameraFrame:
        return CameraFrame.from_request(name, request)

    def capture_frame(
        self, name: str = "main", config=None
    ) -> TypedFuture[CameraFrame]:
        """Make a CameraFrame from the next frame in the named stream.

        :param name: Stream name, defaults to "main"
        :type name: str, optional
        """
        return self._dispatch_loop_tasks(
            LoopTask.with_request(self._capture_frame, name), config=config
        )[0]

    def capture_serial_frames(
        self, n_frames: int, name: str = "main"
    ) -> List[TypedFuture[CameraFrame]]:
        """Capture a number of frames from the named stream, returning a list of CameraFrames."""
        return self._dispatch_loop_tasks(
            *(LoopTask.with_request(self._capture_frame, name) for _ in range(n_frames))
        )
